from dagster.builtins import Any, Bool, Float, Int, Nothing, String
from dagster.config import Enum, EnumValue, Field, Permissive, Selector, Shape
from dagster.config.config_schema import ConfigSchema
from dagster.config.config_type import Array, Noneable, ScalarUnion
from dagster.core.definitions import (
    AssetKey,
    AssetMaterialization,
    CompositeSolidDefinition,
    ConfigMapping,
    DependencyDefinition,
    EventMetadataEntry,
    ExecutorDefinition,
    ExpectationResult,
    Failure,
    FloatMetadataEntryData,
    InputDefinition,
    InputMapping,
    IntMetadataEntryData,
    IntermediateStorageDefinition,
    JsonMetadataEntryData,
    LoggerDefinition,
    MarkdownMetadataEntryData,
    Materialization,
    ModeDefinition,
    MultiDependencyDefinition,
    Output,
    OutputDefinition,
    OutputMapping,
    Partition,
    PartitionSetDefinition,
    PathMetadataEntryData,
    PipelineDefinition,
    PresetDefinition,
    RepositoryDefinition,
    ResourceDefinition,
    RetryRequested,
    ScheduleDefinition,
    ScheduleExecutionContext,
    SolidDefinition,
    SolidInvocation,
    SystemStorageData,
    SystemStorageDefinition,
    TextMetadataEntryData,
    TypeCheck,
    UrlMetadataEntryData,
    composite_solid,
    daily_schedule,
    default_executors,
    executor,
    failure_hook,
    hourly_schedule,
    in_process_executor,
    intermediate_storage,
    lambda_solid,
    logger,
    monthly_schedule,
    multiprocess_executor,
    pipeline,
    reconstructable,
    repository,
    resource,
    schedule,
    solid,
    success_hook,
    system_storage,
    weekly_schedule,
)
from dagster.core.definitions.config_mappable import configured
from dagster.core.errors import (
    DagsterConfigMappingFunctionError,
    DagsterError,
    DagsterEventLogInvalidForRun,
    DagsterExecutionStepExecutionError,
    DagsterExecutionStepNotFoundError,
    DagsterInvalidConfigDefinitionError,
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    DagsterResourceFunctionError,
    DagsterRunNotFoundError,
    DagsterStepOutputNotFoundError,
    DagsterSubprocessError,
    DagsterTypeCheckDidNotPass,
    DagsterTypeCheckError,
    DagsterUnknownResourceError,
    DagsterUnmetExecutorRequirementsError,
    DagsterUserCodeExecutionError,
)
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.execution.api import (
    execute_pipeline,
    execute_pipeline_iterator,
    reexecute_pipeline,
    reexecute_pipeline_iterator,
)
from dagster.core.execution.context.compute import SolidExecutionContext
from dagster.core.execution.context.init import InitResourceContext
from dagster.core.execution.context.logger import InitLoggerContext
from dagster.core.execution.context.system import (
    AssetStoreContext,
    HookContext,
    SystemComputeExecutionContext,
    TypeCheckContext,
)
from dagster.core.execution.results import (
    CompositeSolidExecutionResult,
    PipelineExecutionResult,
    SolidExecutionResult,
)
from dagster.core.executor.base import Executor
from dagster.core.executor.init import InitExecutorContext
from dagster.core.instance import DagsterInstance
from dagster.core.launcher import DefaultRunLauncher
from dagster.core.log_manager import DagsterLogManager
from dagster.core.storage.asset_store import (
    AssetStore,
    custom_path_fs_asset_store,
    fs_asset_store,
    mem_asset_store,
)
from dagster.core.storage.file_manager import FileHandle, LocalFileHandle, local_file_manager
from dagster.core.storage.init import InitIntermediateStorageContext, InitSystemStorageContext
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.storage.system_storage import (
    build_intermediate_storage_from_object_store,
    default_intermediate_storage_defs,
    default_system_storage_defs,
    fs_intermediate_storage,
    fs_system_storage,
    mem_intermediate_storage,
    mem_system_storage,
)
from dagster.core.types.config_schema import (
    dagster_type_loader,
    dagster_type_materializer,
    input_hydration_config,
    output_materialization_config,
)
from dagster.core.types.dagster_type import DagsterType, List, Optional, PythonObjectDagsterType
from dagster.core.types.decorator import (
    make_python_type_usable_as_dagster_type,
    usable_as_dagster_type,
)
from dagster.core.types.marshal import SerializationStrategy
from dagster.core.types.python_dict import Dict
from dagster.core.types.python_set import Set
from dagster.core.types.python_tuple import Tuple
from dagster.utils import file_relative_path
from dagster.utils.test import (
    check_dagster_type,
    execute_solid,
    execute_solid_within_pipeline,
    execute_solids_within_pipeline,
)

from .version import __version__

from dagster.config.source import StringSource, IntSource  # isort:skip


__all__ = [
    # Definition
    "AssetKey",
    "AssetMaterialization",
    "CompositeSolidDefinition",
    "ConfigMapping",
    "DependencyDefinition",
    "EventMetadataEntry",
    "ExecutorDefinition",
    "ExpectationResult",
    "Failure",
    "Field",
    "InputDefinition",
    "InputMapping",
    "IntermediateStorageDefinition",
    "JsonMetadataEntryData",
    "LoggerDefinition",
    "MarkdownMetadataEntryData",
    "IntMetadataEntryData",
    "FloatMetadataEntryData",
    "Materialization",
    "ModeDefinition",
    "MultiDependencyDefinition",
    "Output",
    "OutputDefinition",
    "OutputMapping",
    "PathMetadataEntryData",
    "PipelineDefinition",
    "PresetDefinition",
    "RepositoryDefinition",
    "ResourceDefinition",
    "SolidDefinition",
    "SolidInvocation",
    "SystemStorageDefinition",
    "TextMetadataEntryData",
    "UrlMetadataEntryData",
    # Decorators
    "composite_solid",
    "executor",
    "intermediate_storage",
    "lambda_solid",
    "logger",
    "pipeline",
    "repository",
    "resource",
    "schedule",
    "solid",
    "system_storage",
    "success_hook",
    "failure_hook",
    # Execution
    "CompositeSolidExecutionResult",
    "DagsterEvent",
    "DagsterEventType",
    "DefaultRunLauncher",
    "Executor",
    "InitExecutorContext",
    "InitLoggerContext",
    "InitResourceContext",
    "InitSystemStorageContext",
    "InitIntermediateStorageContext",
    "PipelineExecutionResult",
    "RetryRequested",
    "SolidExecutionResult",
    "SystemComputeExecutionContext",
    "SolidExecutionContext",
    "SystemStorageData",
    "HookContext",
    "TypeCheckContext",
    "AssetStoreContext",
    "PipelineRun",
    "default_executors",
    "default_intermediate_storage_defs",
    "default_system_storage_defs",
    "execute_pipeline_iterator",
    "execute_pipeline",
    "execute_solid_within_pipeline",
    "fs_intermediate_storage",
    "fs_system_storage",
    "in_process_executor",
    "mem_intermediate_storage",
    "mem_system_storage",
    "multiprocess_executor",
    "reconstructable",
    "reexecute_pipeline_iterator",
    "reexecute_pipeline",
    # Errors
    "DagsterConfigMappingFunctionError",
    "DagsterError",
    "DagsterEventLogInvalidForRun",
    "DagsterExecutionStepExecutionError",
    "DagsterExecutionStepNotFoundError",
    "DagsterInvalidConfigDefinitionError",
    "DagsterInvalidConfigError",
    "DagsterInvalidDefinitionError",
    "DagsterInvariantViolationError",
    "DagsterResourceFunctionError",
    "DagsterRunNotFoundError",
    "DagsterStepOutputNotFoundError",
    "DagsterSubprocessError",
    "DagsterTypeCheckDidNotPass",
    "DagsterTypeCheckError",
    "DagsterUnknownResourceError",
    "DagsterUnmetExecutorRequirementsError",
    "DagsterUserCodeExecutionError",
    # Logging
    "DagsterLogManager",
    # Utilities
    "build_intermediate_storage_from_object_store",
    "check_dagster_type",
    "execute_solid",
    "execute_solids_within_pipeline",
    "file_relative_path",
    "configured",
    # types
    "Any",
    "Bool",
    "Dict",
    "Enum",
    "EnumValue",
    "Float",
    "Int",
    "List",
    "Nothing",
    "Optional",
    "SerializationStrategy",
    "Set",
    "String",
    "Tuple",
    "TypeCheck",
    "dagster_type_loader",
    "dagster_type_materializer",
    "input_hydration_config",
    "output_materialization_config",
    # type creation
    "DagsterType",
    "PythonObjectDagsterType",
    "make_python_type_usable_as_dagster_type",
    "usable_as_dagster_type",
    # config
    "Array",
    "ConfigSchema",
    "Noneable",
    "Permissive",
    "ScalarUnion",
    "StringSource",
    "IntSource",
    "Selector",
    "Shape",
    # file things
    "FileHandle",
    "LocalFileHandle",
    "local_file_manager",
    # instance
    "DagsterInstance",
    # partitions and schedules
    "Partition",
    "PartitionSetDefinition",
    "ScheduleDefinition",
    "ScheduleExecutionContext",
    "daily_schedule",
    "hourly_schedule",
    "monthly_schedule",
    "weekly_schedule",
    # asset store
    "AssetStore",
    "mem_asset_store",
    "fs_asset_store",
    "custom_path_fs_asset_store",
]



import sys

# pylint: disable=no-name-in-module,unused-import

# see https://stackoverflow.com/questions/53978542/
if sys.version_info[0] >= 3:
    from collections.abc import (
        Callable,
        Container,
        Hashable,
        ItemsView,
        Iterable,
        Iterator,
        KeysView,
        Mapping,
        MappingView,
        MutableMapping,
        MutableSequence,
        MutableSet,
        Sequence,
        Set,
        Sized,
        ValuesView,
    )
else:
    from collections import (
        Callable,
        Container,
        Hashable,
        ItemsView,
        Iterable,
        Iterator,
        KeysView,
        Mapping,
        MappingView,
        MutableMapping,
        MutableSequence,
        MutableSet,
        Sequence,
        Set,
        Sized,
        ValuesView,
    )

__all__ = [
    "Callable",
    "Container",
    "Hashable",
    "ItemsView",
    "Iterable",
    "Iterator",
    "KeysView",
    "Mapping",
    "MappingView",
    "MutableMapping",
    "MutableSequence",
    "MutableSet",
    "Sequence",
    "Set",
    "Sized",
    "ValuesView",
]



from functools import update_wrapper

from dagster import check
from dagster.builtins import Int
from dagster.config.field import Field
from dagster.config.field_utils import check_user_facing_opt_config_param
from dagster.core.definitions.config_mappable import IConfigMappable
from dagster.core.definitions.reconstructable import ReconstructablePipeline
from dagster.core.errors import DagsterUnmetExecutorRequirementsError
from dagster.core.execution.retries import Retries, get_retries_config
from dagster.utils.backcompat import rename_warning


class ExecutorDefinition(IConfigMappable):
    """
    Args:
        name (Optional[str]): The name of the executor.
        config_schema (Optional[ConfigSchema]): The schema for the config. Configuration data
            available in `init_context.executor_config`.
        executor_creation_fn(Optional[Callable]): Should accept an :py:class:`InitExecutorContext`
            and return an instance of :py:class:`Executor`
        required_resource_keys (Optional[Set[str]]): Keys for the resources required by the
            executor.
        _configured_config_mapping_fn: This argument is for internal use only. Users should not
            specify this field. To preconfigure a resource, use the :py:func:`configured` API.
    """

    def __init__(
        self,
        name,
        config_schema=None,
        executor_creation_fn=None,
        required_resource_keys=None,
        _configured_config_mapping_fn=None,
    ):
        self._name = check.str_param(name, "name")
        self._config_schema = check_user_facing_opt_config_param(config_schema, "config_schema")
        self._executor_creation_fn = check.opt_callable_param(
            executor_creation_fn, "executor_creation_fn"
        )
        self._required_resource_keys = frozenset(
            check.opt_set_param(required_resource_keys, "required_resource_keys", of_type=str)
        )
        self.__configured_config_mapping_fn = check.opt_callable_param(
            _configured_config_mapping_fn, "config_mapping_fn"
        )

    @property
    def name(self):
        return self._name

    @property
    def config_field(self):
        rename_warning("config_schema", "config_field", "0.9.0")
        return self._config_schema

    @property
    def config_schema(self):
        return self._config_schema

    @property
    def executor_creation_fn(self):
        return self._executor_creation_fn

    @property
    def required_resource_keys(self):
        return self._required_resource_keys

    @property
    def _configured_config_mapping_fn(self):
        return self.__configured_config_mapping_fn

    def configured(self, config_or_config_fn, config_schema=None, **kwargs):
        wrapped_config_mapping_fn = self._get_wrapped_config_mapping_fn(
            config_or_config_fn, config_schema
        )

        return ExecutorDefinition(
            config_schema=config_schema,
            name=kwargs.get("name", self.name),
            executor_creation_fn=self.executor_creation_fn,
            required_resource_keys=self.required_resource_keys,
            _configured_config_mapping_fn=wrapped_config_mapping_fn,
        )


def executor(name=None, config_schema=None, required_resource_keys=None):
    """Define an executor.

    The decorated function should accept an :py:class:`InitExecutorContext` and return an instance
    of :py:class:`Executor`.

    Args:
        name (Optional[str]): The name of the executor.
        config_schema (Optional[ConfigSchema]): The schema for the config. Configuration data available in
            `init_context.executor_config`.
        required_resource_keys (Optional[Set[str]]): Keys for the resources required by the
            executor.
    """
    if callable(name):
        check.invariant(config_schema is None)
        check.invariant(required_resource_keys is None)
        return _ExecutorDecoratorCallable()(name)

    return _ExecutorDecoratorCallable(
        name=name, config_schema=config_schema, required_resource_keys=required_resource_keys,
    )


class _ExecutorDecoratorCallable(object):
    def __init__(self, name=None, config_schema=None, required_resource_keys=None):
        self.name = check.opt_str_param(name, "name")
        self.config_schema = config_schema  # type check in definition
        self.required_resource_keys = required_resource_keys  # type check in definition

    def __call__(self, fn):
        check.callable_param(fn, "fn")

        if not self.name:
            self.name = fn.__name__

        executor_def = ExecutorDefinition(
            name=self.name,
            config_schema=self.config_schema,
            executor_creation_fn=fn,
            required_resource_keys=self.required_resource_keys,
        )

        update_wrapper(executor_def, wrapped=fn)

        return executor_def


@executor(
    name="in_process",
    config_schema={
        "retries": get_retries_config(),
        "marker_to_close": Field(str, is_required=False),
    },
)
def in_process_executor(init_context):
    """The default in-process executor.

    In most Dagster environments, this will be the default executor. It is available by default on
    any :py:class:`ModeDefinition` that does not provide custom executors. To select it explicitly,
    include the following top-level fragment in config:

    .. code-block:: yaml

        execution:
          in_process:

    Execution priority can be configured using the ``dagster/priority`` tag via solid metadata,
    where the higher the number the higher the priority. 0 is the default and both positive
    and negative numbers can be used.
    """
    from dagster.core.executor.init import InitExecutorContext
    from dagster.core.executor.in_process import InProcessExecutor

    check.inst_param(init_context, "init_context", InitExecutorContext)

    return InProcessExecutor(
        # shouldn't need to .get() here - issue with defaults in config setup
        retries=Retries.from_config(init_context.executor_config.get("retries", {"enabled": {}})),
        marker_to_close=init_context.executor_config.get("marker_to_close"),
    )


@executor(
    name="multiprocess",
    config_schema={
        "max_concurrent": Field(Int, is_required=False, default_value=0),
        "retries": get_retries_config(),
    },
)
def multiprocess_executor(init_context):
    """The default multiprocess executor.

    This simple multiprocess executor is available by default on any :py:class:`ModeDefinition`
    that does not provide custom executors. To select the multiprocess executor, include a fragment
    such as the following in your config:

    .. code-block:: yaml

        execution:
          multiprocess:
            config:
              max_concurrent: 4

    The ``max_concurrent`` arg is optional and tells the execution engine how many processes may run
    concurrently. By default, or if you set ``max_concurrent`` to be 0, this is the return value of
    :py:func:`python:multiprocessing.cpu_count`.

    Execution priority can be configured using the ``dagster/priority`` tag via solid metadata,
    where the higher the number the higher the priority. 0 is the default and both positive
    and negative numbers can be used.
    """
    from dagster.core.executor.init import InitExecutorContext
    from dagster.core.executor.multiprocess import MultiprocessExecutor

    check.inst_param(init_context, "init_context", InitExecutorContext)

    check_cross_process_constraints(init_context)

    return MultiprocessExecutor(
        pipeline=init_context.pipeline,
        max_concurrent=init_context.executor_config["max_concurrent"],
        retries=Retries.from_config(init_context.executor_config["retries"]),
    )


default_executors = [in_process_executor, multiprocess_executor]


def check_cross_process_constraints(init_context):
    from dagster.core.executor.init import InitExecutorContext

    check.inst_param(init_context, "init_context", InitExecutorContext)

    _check_intra_process_pipeline(init_context.pipeline)
    _check_non_ephemeral_instance(init_context.instance)
    _check_persistent_storage_requirement(
        init_context.intermediate_storage_def, init_context.system_storage_def
    )


def _check_intra_process_pipeline(pipeline):
    if not isinstance(pipeline, ReconstructablePipeline):
        raise DagsterUnmetExecutorRequirementsError(
            'You have attempted to use an executor that uses multiple processes with the pipeline "{name}" '
            "that is not reconstructable. Pipelines must be loaded in a way that allows dagster to reconstruct "
            "them in a new process. This means: \n"
            "  * using the file, module, or repository.yaml arguments of dagit/dagster-graphql/dagster\n"
            "  * loading the pipeline through the reconstructable() function\n".format(
                name=pipeline.get_definition().name
            )
        )


def _check_persistent_storage_requirement(intermediate_storage_def, system_storage_def):
    if intermediate_storage_def:
        if not intermediate_storage_def.is_persistent:
            raise DagsterUnmetExecutorRequirementsError(
                (
                    "You have attempted to use an executor that uses multiple processes while using "
                    "intermediate storage {storage_name} which does not persist intermediates. "
                    "This means there would be no way to move data between different "
                    "processes. Please configure your pipeline in the storage config "
                    "section to use persistent system storage such as the filesystem."
                ).format(storage_name=intermediate_storage_def.name)
            )
    elif not system_storage_def.is_persistent:
        raise DagsterUnmetExecutorRequirementsError(
            (
                "You have attempted to use an executor that uses multiple processes while using system "
                "storage {storage_name} which does not persist intermediates. "
                "This means there would be no way to move data between different "
                "processes. Please configure your pipeline in the storage config "
                "section to use persistent system storage such as the filesystem."
            ).format(storage_name=system_storage_def.name)
        )


def _check_non_ephemeral_instance(instance):
    if instance.is_ephemeral:
        raise DagsterUnmetExecutorRequirementsError(
            "You have attempted to use an executor that uses multiple processes with an "
            "ephemeral DagsterInstance. A non-ephemeral instance is needed to coordinate "
            "execution between multiple processes. You can configure your default instance "
            "via $DAGSTER_HOME or ensure a valid one is passed when invoking the python APIs."
        )



import threading

import grpc
from dagster import check
from dagster.grpc.client import DagsterGrpcClient

WATCH_INTERVAL = 1
MAX_RECONNECT_ATTEMPTS = 10


def watch_grpc_server_thread(
    client,
    on_disconnect,
    on_reconnected,
    on_updated,
    on_error,
    shutdown_event,
    watch_interval=None,
    max_reconnect_attempts=None,
):

    """
    This thread watches the state of the unmanaged gRPC server and calls the appropriate handler
    functions in case of a change.

    The following loop polls the GetServerId endpoint to check if either:
    1. The server_id has changed
    2. The server is unreachable

    In the case of (1) The server ID has changed, we call `on_updated` and end the thread.

    In the case of (2) The server is unreachable, we attempt to automatically reconnect. If we
    are able to reconnect, there are two possibilities:

    a. The server ID has changed
        -> In this case, we we call `on_updated` and end the thread.
    b. The server ID is the same
        -> In this case, we we call `on_reconnected`, and we go back to polling the server for
        changes.

    If we are unable to reconnect to the server within the specified max_reconnect_attempts, we
    call on_error.

    Once the on_updated or on_error events are called, this thread shuts down completely. These two
    events are called at most once, while `on_disconnected` and `on_reconnected` may be called
    multiple times in order to be properly handle intermittent network failures.
    """

    check.inst_param(client, "client", DagsterGrpcClient)
    check.callable_param(on_disconnect, "on_disconnect")
    check.callable_param(on_reconnected, "on_reconnected")
    check.callable_param(on_updated, "on_updated")
    check.callable_param(on_error, "on_error")
    watch_interval = check.opt_numeric_param(watch_interval, "watch_interval", WATCH_INTERVAL)
    max_reconnect_attempts = check.opt_int_param(
        max_reconnect_attempts, "max_reconnect_attempts", MAX_RECONNECT_ATTEMPTS
    )

    server_id = {"current": None}

    def current_server_id():
        return server_id["current"]

    def set_server_id(new_id):
        server_id["current"] = new_id

    def watch_for_changes():
        while True:
            if shutdown_event.is_set():
                break

            curr = current_server_id()
            new_server_id = client.get_server_id()
            if curr is None:
                set_server_id(new_server_id)
            elif curr != new_server_id:
                set_server_id(new_server_id)
                return on_updated()

            shutdown_event.wait(watch_interval)

    def reconnect_loop():
        attempts = 0
        while attempts < max_reconnect_attempts:
            shutdown_event.wait(watch_interval)
            if shutdown_event.is_set():
                return False

            try:
                new_server_id = client.get_server_id()
                if current_server_id() == new_server_id:
                    # Intermittent failure, was able to reconnect to the same server
                    on_reconnected()
                    return True
                else:
                    on_updated()
                    set_server_id(new_server_id)
                    return False
            except grpc.RpcError:  # pylint: disable=protected-access
                attempts += 1

        on_error()
        return False

    while True:
        try:
            watch_for_changes()
            return
        except grpc.RpcError:  # pylint: disable=protected-access
            on_disconnect()
            reconnected_to_same_server = reconnect_loop()
            if not reconnected_to_same_server:
                return


def create_grpc_watch_thread(
    client,
    on_disconnect=None,
    on_reconnected=None,
    on_updated=None,
    on_error=None,
    watch_interval=None,
    max_reconnect_attempts=None,
):
    check.inst_param(client, "client", DagsterGrpcClient)

    noop = lambda *a: None
    on_disconnect = check.opt_callable_param(on_disconnect, "on_disconnect", noop)
    on_reconnected = check.opt_callable_param(on_reconnected, "on_reconnected", noop)
    on_updated = check.opt_callable_param(on_updated, "on_updated", noop)
    on_error = check.opt_callable_param(on_error, "on_error", noop)

    shutdown_event = threading.Event()
    thread = threading.Thread(
        target=watch_grpc_server_thread,
        args=[
            client,
            on_disconnect,
            on_reconnected,
            on_updated,
            on_error,
            shutdown_event,
            watch_interval,
            max_reconnect_attempts,
        ],
    )
    thread.daemon = True
    return shutdown_event, thread
